@@ -2,12 +2,7 @@
  * built-in Korn commands: c_*
  */
 
-#if !defined(lint) && !defined(no_RCSids)
-static char *RCSid = "$Id: c_ksh.c,v 1.3 1994/05/31 13:34:34 michael Exp $";
-#endif
-
 #include "sh.h"
-#include "expand.h"
 #include <ctype.h>
 
 int
@@ -23,16 +18,24 @@ c_cd(wp)
 	extern int	make_path();
 	extern void	simplify_path();
 
+	if (Flag(FRESTRICTED)) {
+		bi_errorf("restricted shell - can't cd");
+		return 1;
+	}
 	pwd = strval(pwd_s = global("PWD"));
 	oldpwd = strval(oldpwd_s = global("OLDPWD"));
 
 	if ((cp = wp[1]) == (char *) 0) {
-		if ((cp = strval(global("HOME"))) == null)
-			errorf("cd: no home directory\n");
+		if ((cp = strval(global("HOME"))) == null) {
+			bi_errorf("no home directory");
+			return 1;
+		}
 	} else if (wp[2] == (char *) 0) {
 		if (strcmp(cp, "-") == 0) {
-			if ((cp = oldpwd) == null)
-				errorf("cd: no OLDPWD\n");
+			if ((cp = oldpwd) == null) {
+				bi_errorf("no OLDPWD");
+				return 1;
+			}
 			printpath++;
 		}
 	} else if (wp[3] == (char *) 0) {
@@ -43,21 +46,28 @@ c_cd(wp)
 		 * we could try to find another substitution. For now
 		 * we don't
 		 */
-		if ((cp = strstr(pwd, wp[1])) == (char *) 0)
-			errorf("cd: bad substitution\n");
+		if ((cp = strstr(pwd, wp[1])) == (char *) 0) {
+			bi_errorf("bad substitution");
+			return 1;
+		}
 		ilen = cp - pwd;
 		olen = strlen(wp[1]);
 		nlen = strlen(wp[2]);
 		if (ilen + nlen + strlen(pwd + ilen + olen) + 1 > sizeof(path))
-			errorf("cd: path too long\n");
+		{
+			bi_errorf("path too long");
+			return 1;
+		}
 		strncpy(path, pwd, ilen);
 		strcpy(path + ilen, wp[2]);
 		strcpy(path + ilen + nlen, pwd + ilen + olen);
 		cp = path;
 		printpath++;
 		/* XXX: make_path() copies cp to path, in this case cp to cp. */
-	} else
-		errorf("cd: too many arguments\n");
+	} else {
+		bi_errorf("too many arguments");
+		return 1;
+	}
 
 	cdpath = strval(global("CDPATH"));
 	do {
@@ -66,13 +76,24 @@ c_cd(wp)
 		rval = chdir(path);
 	} while (rval < 0 && cdpath != (char *) 0);
 
-	if (rval < 0)
-		errorf("cd: %s: bad directory\n", cp);
+	if (rval < 0) {
+		bi_errorf("%s: bad directory", cp);
+		return 1;
+	}
 
 	flushcom(0);
 
 	if (printpath || cdnode)
-		shellf("%s\n", path);
+		shprintf("%s\n", path);
+
+#ifdef OS2
+	/* OS2 is like MSDOS in that paths can be relative to a drive
+	 * specification, hence, the above code doesn't always calculate
+	 * the current directory correctly.  Workaround is to let getcwd()
+	 * figure it out.
+	 */
+	getcwd(path, sizeof(path));
+#endif /* OS2 */
 
 	setstr(oldpwd_s, pwd);
 	setstr(pwd_s, path);
@@ -88,44 +109,84 @@ c_print(wp)
 #define PO_EXPAND	BIT(1)	/* expand backslash sequences */
 #define PO_ECHO		BIT(2)	/* act (kind of) like BSD echo */
 #define PO_HIST		BIT(3)	/* print to history instead of stdout */
+#define PO_COPROC	BIT(4)	/* printing to coprocess: block SIGPIPE */
 	int fd = 1;
 	int flags = PO_EXPAND|PO_NL;
 	char *s;
+	char *emsg;
 	XString xs;
 	char *xp;
 
-	if (wp[0][0] == 'e')	/* strict sysV echo: escapes, no options */
+	if (wp[0][0] == 'e') {	/* echo command */
+		int nflags = flags;
+
+		/* A compromise between sysV and BSD echo commands:
+		 * escape sequences are enabled by default, and
+		 * -n, -e and -E are recognized if they appear
+		 * in arguments with no illegal options (ie, echo -nq
+		 * will print -nq).
+		 * Different from sysV echo since options are recognized,
+		 * different from BSD echo since escape sequences are enabled
+		 * by default.
+		 */
 		wp += 1;
-	else {
-		int optc, fl;
-		char *options = "Rnrsu,";
+		while ((s = *wp) && *s == '-') {
+			while (*++s)
+				if (*s == 'n')
+					nflags &= ~PO_NL;
+				else if (*s == 'e')
+					nflags |= PO_EXPAND;
+				else if (*s == 'E')
+					nflags &= ~PO_EXPAND;
+				else
+					/* bad option: don't use nflags, print
+					 * argument
+					 */
+					break;
+			if (*s)
+				break;
+			wp++;
+			flags = nflags;
+		}
+	} else {
+		int optc;
+		char *options = "Rnprsu,";
 
 		while ((optc = ksh_getopt(wp, &builtin_opt, options)) != EOF)
 			switch (optc) {
-			case 'R':
+			  case 'R':
 				flags |= PO_ECHO; /* fake BSD echo command */
 				flags &= ~PO_EXPAND;
 				options = "ne";
 				break;
-			case 'e':
+			  case 'e':
 				flags |= PO_EXPAND;
 				break;
-			case 'n':
+			  case 'n':
 				flags &= ~PO_NL;
 				break;
-			case 'r':
+			  case 'p':
+				if ((fd = get_coproc_fd(W_OK, &emsg)) < 0) {
+					bi_errorf("-p: %s", emsg);
+					return 1;
+				}
+				break;
+			  case 'r':
 				flags &= ~PO_EXPAND;
 				break;
-			case 's':
+			  case 's':
 				flags |= PO_HIST;
 				break;
-			case 'u':
-				s = builtin_opt.optarg;
-				if (*s
-				    && (!digit(*s)
-				      || (fl = fcntl(fd = *s++ - '0', F_GETFL, 0) < 0)
-				      || (fl != O_WRONLY && fl != O_RDWR)))
-					errorf("print: bad -u argument\n");
+			  case 'u':
+				if (!*(s = builtin_opt.optarg))
+					fd = 0;
+				else if ((fd = check_fd(s, W_OK, &emsg)) < 0) {
+					bi_errorf("-u: %s: %s", s, emsg);
+					return 1;
+				}
+				break;
+			  case '?':
+				return 1;
 			}
 		if (!(builtin_opt.info & GI_MINUSMINUS)) {
 			/* treat a lone - like -- */
@@ -137,7 +198,7 @@ c_print(wp)
 		wp += builtin_opt.optind;
 	}
 
-	Xinit(xs, xp, 128);
+	Xinit(xs, xp, 128, ATEMP);
 
 	while (*wp != NULL) {
 		register int c;
@@ -183,44 +244,63 @@ c_print(wp)
 
 	if (flags & PO_HIST) {
 		Xput(xs, xp, '\0');
-#ifdef EASY_HISTORY
-		histsave(Xstring(xs, xp));
-#else /* EASY_HISTORY */
-		/* todo: is incremented line here correct? */
-		histsave(++(source->line), Xstring(xs, xp), 1);
-#endif /* EASY_HISTORY */
+		source->line++;
+		histsave(source->line, Xstring(xs, xp), 1);
 		Xfree(xs, xp);
 	} else {
 		int n, len = Xlength(xs, xp);
+		int UNINITIALIZED(opipe);
 
+		/* Ensure we aren't killed by a SIGPIPE while writing to
+		 * a coprocess.  at&t ksh doesn't seem to do this (seems
+		 * to just check that the co-process is alive, which is
+		 * not enough).
+		 */
+		if (coproc.write >= 0 && coproc.write == fd) {
+			flags |= PO_COPROC;
+			opipe = block_pipe();
+		}
 		for (s = Xstring(xs, xp); len > 0; ) {
 			n = write(fd, s, len);
 			if (n < 0) {
+				if (flags & PO_COPROC)
+					restore_pipe(opipe);
 				if (errno == EINTR) {
+					/* allow user to ^C out */
 					intrcheck();
+					if (flags & PO_COPROC)
+						opipe = block_pipe();
 					continue;
 				}
+				if (errno == EPIPE)
+					coproc_write_close(fd);
 				return 1;
 			}
 			s += n;
 			len -= n;
 		}
+		if (flags & PO_COPROC)
+			restore_pipe(opipe);
 	}
 
 	return 0;
 }
 
+#ifdef KSH
 int
 c_whence(wp)
 	register char **wp;
 {
-	register struct tbl *tp;
+	struct tbl *tp;
 	char *id;
-	int vflag = 0, pflag = 0;
+	int pflag = 0, vflag = 0, Vflag = 0;
 	int ret = 0;
 	int optc;
+	int iam_whence = wp[0][0] == 'w';
+	int fcflags;
+	char *options = iam_whence ? "pv" : "pvV";
 
-	while ((optc = ksh_getopt(wp, &builtin_opt, "pv")) != EOF)
+	while ((optc = ksh_getopt(wp, &builtin_opt, options)) != EOF)
 		switch (optc) {
 		case 'p':
 			pflag = 1;
@@ -228,22 +308,40 @@ c_whence(wp)
 		case 'v':
 			vflag = 1;
 			break;
+		case 'V':
+			Vflag = 1;
+			break;
+		case '?':
+			return 1;
 		}
 	wp += builtin_opt.optind;
 
-	while ((vflag || ret == 0) && (id = *wp++) != NULL) {
+
+	fcflags = FC_BI | FC_PATH | FC_FUNC;
+	if (!iam_whence) {
+		/* Note that -p on its own is deal with in comexec() */
 		if (pflag)
-			tp = NULL;
-		else {
+			fcflags |= FC_DEFPATH;
+		/* Convert command options to whence options - note that
+		 * command -pV uses a different path search than whence -v
+		 * or whence -pv.  This should be considered a feature.
+		 */
+		vflag = Vflag;
+	}
+	if (pflag)
+		fcflags &= ~(FC_BI | FC_FUNC);
+
+	while ((vflag || ret == 0) && (id = *wp++) != NULL) {
+		tp = NULL;
+		if (iam_whence && !pflag)
 			tp = tsearch(&keywords, id, hash(id));
-			if (tp == NULL) {
-				tp = tsearch(&aliases, id, hash(id));
-				if (tp && !(tp->flag & ISSET))
-					tp = NULL;
-			}
+		if (!tp && !pflag) {
+			tp = tsearch(&aliases, id, hash(id));
+			if (tp && !(tp->flag & ISSET))
+				tp = NULL;
 		}
-		if (tp == NULL)
-			tp = findcom(id, Flag(FTRACKALL), FALSE, pflag);
+		if (!tp)
+			tp = findcom(id, fcflags);
 		if (vflag || (tp->type != CALIAS && tp->type != CEXEC
 			      && tp->type != CTALIAS))
 			shprintf("%s", id);
@@ -254,12 +352,12 @@ c_whence(wp)
 			break;
 		  case CALIAS:
 			if (vflag)
-				/* odd but at&t ksh doesn't use quoting here */
-				shprintf(" is an %salias for %s",
-					(tp->flag & EXPORT) ? "exported " : "",
-					tp->val.s);
-			else
-				print_value_quoted(tp->val.s);
+				shprintf(" is an %salias for ",
+					(tp->flag & EXPORT) ? "exported "
+							    : null);
+			if (!iam_whence && !vflag)
+				shprintf("alias %s=", id);
+			print_value_quoted(tp->val.s);
 			break;
 		  case CFUNC:
 			if (vflag) {
@@ -268,14 +366,19 @@ c_whence(wp)
 					shprintf(" exported");
 				if (tp->flag & TRACE)
 					shprintf(" traced");
-				if (!(tp->flag & ISSET))
+				if (!(tp->flag & ISSET)) {
 					shprintf(" undefined");
+					if (tp->u.fpath)
+						shprintf(" (autoload from %s)",
+							tp->u.fpath);
+				}
 				shprintf(" function");
 			}
 			break;
 		  case CSHELL:
 			if (vflag)
-				shprintf(" is a shell builtin");
+				shprintf(" is a%s shell builtin",
+				    (tp->flag & SPEC_BI) ? " special" : null);
 			break;
 		  case CTALIAS:
 		  case CEXEC:
@@ -287,7 +390,7 @@ c_whence(wp)
 						    "a tracked %salias for ",
 							(tp->flag & EXPORT) ?
 								"exported "
-							      : "");
+							      : null);
 				}
 				shprintf("%s", tp->val.s);
 			} else {
@@ -301,10 +404,31 @@ c_whence(wp)
 			break;
 		}
 		if (vflag || !ret)
-			shprintf("\n");
+			shprintf(newline);
 	}
 	return ret;
 }
+#endif /* KSH */
+
+/* Deal with command -vV - command -p dealt with in comexec() */
+int
+c_command(wp)
+	register char **wp;
+{
+#ifdef KSH
+	/* Let c_whence do the work.  Note that c_command() must be
+	 * a distinct function from c_whence() (tested in comexec()).
+	 */
+	return c_whence(wp);
+#else /* KSH */
+	/* Can't get here: command with no options taken care of in
+	 * comexec().
+	 */
+	if (ksh_getopt(wp, &builtin_opt, null) == '?')
+		return 1;
+	return 0;
+#endif /* KSH */
+}
 
 /* typeset, export, and readonly */
 int
@@ -395,6 +519,8 @@ c_typeset(wp)
 		  case 'x':
 			flag |= EXPORT;
 			break;
+		  case '?':
+			return 1;
 		}
 		if (builtin_opt.info & GI_PLUS) {
 			fclr |= flag;
@@ -407,8 +533,12 @@ c_typeset(wp)
 		}
 	}
 
-	field = fieldstr ? getn(fieldstr) : 0;
-	base = basestr ? getn(basestr) : 0;
+	field = 0;
+	if (fieldstr && !bi_getn(fieldstr, &field))
+		return 1;
+	base = 0;
+	if (basestr && !bi_getn(basestr, &base))
+		return 1;
 
 	if (!(builtin_opt.info & GI_MINUSMINUS) && wp[builtin_opt.optind]
 	    && (wp[builtin_opt.optind][0] == '-'
@@ -419,10 +549,10 @@ c_typeset(wp)
 		builtin_opt.optind++;
 	}
 
-	if (func && ((fset|fclr) & ~(TRACE|UCASEV_AL|EXPORT)))
-		errorf("%s: only -t, -u and -x options may be used with -f\n",
-			wp[0]);
-	else if (wp[builtin_opt.optind]) {
+	if (func && ((fset|fclr) & ~(TRACE|UCASEV_AL|EXPORT))) {
+		bi_errorf("only -t, -u and -x options may be used with -f");
+		return 1;
+	} else if (wp[builtin_opt.optind]) {
 		/* Take care of exclusions */
 		/* setting these attributes clears the others, unless they
 		 * are also set in this command
@@ -457,6 +587,7 @@ c_typeset(wp)
 	/* set variables and attributes */
 	if (wp[builtin_opt.optind]) {
 		int i;
+		int rval = 0;
 		struct tbl *f;
 
 		if (local && !func)
@@ -465,26 +596,31 @@ c_typeset(wp)
 			if (func) {
 				f = findfunc(wp[i], hash(wp[i]),
 					     (fset&UCASEV_AL) ? TRUE : FALSE);
-				if (!f)
+				if (!f) {
+					/* at&t ksh does ++rval: bogus */
+					rval = 1;
 					continue;
+				}
 				if (fset | fclr) {
 					f->flag |= fset;
 					f->flag &= ~fclr;
 				} else
-					fptreef(shl_stdout, "function %s %T\n",
+					fptreef(shl_stdout, 0,
+						"function %s %T\n",
 						wp[i], f->val.t);
-			} else if (!typeset(wp[i], fset, fclr, field, base))
-				errorf("%s: %s: not identifier\n",
-					wp[0], wp[i]);
+			} else if (!typeset(wp[i], fset, fclr, field, base)) {
+				bi_errorf("%s: not identifier", wp[i]);
+				return 1;
+			}
 		}
-		return 0;
+		return rval;
 	}
 
 	/* list variables and attributes */
 	flag = fset | fclr; /* no difference at this point.. */
 	for (l = e->loc; l; l = l->next) {
 	    for (p = tsort(func ? &l->funs : &l->vars); (vp = *p++); )
-		for (; vp; vp = vp->array) {
+		for (; vp; vp = vp->u.array) {
 		    if (!(vp->flag&ISSET))
 			continue;
 		    /* no arguments */
@@ -523,7 +659,7 @@ c_typeset(wp)
 				continue;
 			if (func) {
 			    if (thing == '-')
-				fptreef(shl_stdout, "function %s %T\n",
+				fptreef(shl_stdout, 0, "function %s %T\n",
 					vp->name, vp->val.t);
 			    else
 				shprintf("%s\n", vp->name);
@@ -546,7 +682,7 @@ c_typeset(wp)
 				else
 				    print_value_quoted(s);
 			    }
-			    shprintf("\n");
+			    shprintf(newline);
 			}
 		    }
 		}
@@ -559,14 +695,17 @@ c_alias(wp)
 	register char **wp;
 {
 	struct table *t = &aliases;
-	int rv = 0, tflag, Uflag = 0, xflag = 0;
+	int rv = 0, rflag = 0, tflag, Uflag = 0, xflag = 0;
 	int optc;
 
-	while ((optc = ksh_getopt(wp, &builtin_opt, "dtUx")) != EOF)
+	while ((optc = ksh_getopt(wp, &builtin_opt, "drtUx")) != EOF)
 		switch (optc) {
 		  case 'd':
 			t = &homedirs;
 			break;
+		  case 'r':
+			rflag = 1;
+			break;
 		  case 't':
 			t = &taliases;
 			break;
@@ -578,9 +717,25 @@ c_alias(wp)
 		  case 'x':
 			xflag = EXPORT;
 			break;
+		  case '?':
+			return 1;
 		}
 	wp += builtin_opt.optind;
 
+	tflag = t == &taliases;
+
+	/* "hash -r" means reset all the tracked aliases.. */
+	if (rflag) {
+		static char *args[] = { "unalias", "-ta", (char *) 0 };
+
+		if (!tflag || *wp) {
+			shprintf(
+	    "alias: -r flag can only be used with -t and without arguments\n");
+			return 1;
+		}
+		return c_unalias(args);
+	}
+
 	if (*wp == NULL) {
 		struct tbl *ap, **p;
 
@@ -588,12 +743,10 @@ c_alias(wp)
 			if ((ap->flag & (ISSET|xflag)) == (ISSET|xflag)) {
 				shprintf("%s=", ap->name);
 				print_value_quoted(ap->val.s);
-				shprintf("\n");
+				shprintf(newline);
 			}
 	}
 
-	tflag = t == &taliases;
-
 	for (; *wp != NULL; wp++) {
 		char *alias = *wp;
 		char *val = strchr(alias, '=');
@@ -609,7 +762,7 @@ c_alias(wp)
 			if (ap != NULL && (ap->flag&ISSET)) {
 				shprintf("%s=", ap->name);
 				print_value_quoted(ap->val.s);
-				shprintf("\n");
+				shprintf(newline);
 			} else {
 				shprintf("%s alias not found\n", alias);
 				rv = 1;
@@ -660,6 +813,8 @@ c_unalias(wp)
 		  case 't':
 			t = &taliases;
 			break;
+		  case '?':
+			return 1;
 		}
 	wp += builtin_opt.optind;
 
@@ -696,9 +851,10 @@ c_let(wp)
 	int rv = 1;
 
 	if (wp[1] == (char *) 0) /* at&t ksh does this */
-		errorf("%s: no arguments\n");
-	for (wp++; *wp; wp++)
-		rv = evaluate(*wp) == 0;
+		bi_errorf("no arguments");
+	else
+		for (wp++; *wp; wp++)
+			rv = evaluate(*wp) == 0;
 	return rv;
 }
 
@@ -709,29 +865,34 @@ c_jobs(wp)
 	int optc;
 	int flag = 0;
 	int nflag = 0;
+	int rv = 0;
 
 	while ((optc = ksh_getopt(wp, &builtin_opt, "lpnz")) != EOF)
 		switch (optc) {
-		case 'l':
+		  case 'l':
 			flag = 1;
 			break;
-		case 'p':
+		  case 'p':
 			flag = 2;
 			break;
-		case 'n':
+		  case 'n':
 			nflag = 1;
 			break;
-		case 'z':	/* debugging: print zombies */
+		  case 'z':	/* debugging: print zombies */
 			nflag = -1;
 			break;
+		  case '?':
+			return 1;
 		}
 	wp += builtin_opt.optind;
 	if (!*wp)
-		j_jobs((char *) 0, flag, nflag);
+		if (j_jobs((char *) 0, flag, nflag))
+			rv = 1;
 	else
 		for (; *wp; wp++)
-			j_jobs(*wp, flag, nflag);
-	return 0;
+			if (j_jobs(*wp, flag, nflag))
+				rv = 1;
+	return rv;
 }
 
 #ifdef JOBS
@@ -742,9 +903,12 @@ c_fgbg(wp)
 	int bg = strcmp(*wp, "bg") == 0;
 	int UNINITIALIZED(rv);
 
-	if (!Flag(FMONITOR))
-		errorf("%s: Job control not enabled\n", *wp);
-	ksh_getopt(wp, &builtin_opt, "");
+	if (!Flag(FMONITOR)) {
+		bi_errorf("job control not enabled\n");
+		return 1;
+	}
+	if (ksh_getopt(wp, &builtin_opt, null) == '?')
+		return 1;
 	wp += builtin_opt.optind;
 	if (*wp)
 		for (; *wp; wp++)
@@ -758,6 +922,36 @@ c_fgbg(wp)
 }
 #endif
 
+struct kill_info {
+	int num_width;
+	int name_width;
+};
+
+/* format a single kill item */
+static char *
+kill_fmt_entry(arg, i, buf, buflen)
+	void *arg;
+	int i;
+	char *buf;
+	int buflen;
+{
+	struct kill_info *ki = (struct kill_info *) arg;
+
+	i++;
+	if (sigtraps[i].name)
+		shf_snprintf(buf, buflen, "%*d %*s %s",
+			ki->num_width, i,
+			ki->name_width, sigtraps[i].name,
+			sigtraps[i].mess);
+	else
+		shf_snprintf(buf, buflen, "%*d %*d %s",
+			ki->num_width, i,
+			ki->name_width, sigtraps[i].signal,
+			sigtraps[i].mess);
+	return buf;
+}
+
+
 int
 c_kill(wp)
 	register char **wp;
@@ -769,34 +963,44 @@ c_kill(wp)
 
 	/* assume old style options if -digits or -UPPERCASE */
 	if ((p = wp[1]) && *p == '-' && (digit(p[1]) || isupper(p[1]))) {
-		if (!(t = gettrap(p + 1)))
-			errorf("kill: bad signal `%s'\n", p + 1);
+		if (!(t = gettrap(p + 1))) {
+			bi_errorf("bad signal `%s'", p + 1);
+			return 1;
+		}
 		i = (wp[2] && strcmp(wp[2], "--") == 0) ? 3 : 2;
 	} else {
 		int optc;
 
 		while ((optc = ksh_getopt(wp, &builtin_opt, "ls:")) != EOF)
 			switch (optc) {
-			case 'l':
+			  case 'l':
 				lflag = 1;
 				break;
-			case 's':
-				if (!(t = gettrap(builtin_opt.optarg)))
-					errorf("kill: bad signal `%s'\n",
+			  case 's':
+				if (!(t = gettrap(builtin_opt.optarg))) {
+					bi_errorf("bad signal `%s'",
 						builtin_opt.optarg);
+					return 1;
+				}
+			  case '?':
+				return 1;
 			}
 		i = builtin_opt.optind;
 	}
-	if ((lflag && t) || (!wp[i] && !lflag))
-		errorf(
-"Usage: kill [ -s signal | -signum | -NAME ] {pid|job}...\n\
+	if ((lflag && t) || (!wp[i] && !lflag)) {
+		shf_fprintf(shl_out,
+"Usage: kill [ -s signame | -signum | -signame ] {pid|job}...\n\
        kill -l [exit_status]\n"
 			);
+		bi_errorf((char *) 0);
+		return 1;
+	}
 
 	if (lflag) {
 		if (wp[i]) {
 			for (; wp[i]; i++) {
-				n = getn(wp[i]);
+				if (!bi_getn(wp[i], &n))
+					return 1;
 				if (n > 128 && n < 128 + SIGNALS)
 					n -= 128;
 				if (n > 0 && n < SIGNALS && sigtraps[n].name)
@@ -805,34 +1009,32 @@ c_kill(wp)
 					shprintf("%d\n", n);
 			}
 		} else if (Flag(FPOSIX)) {
-			p = "";
-			for (i = 1; i < SIGNALS; i++, p = " ")
+			p = null;
+			for (i = 1; i < SIGNALS; i++, p = space)
 				if (sigtraps[i].name)
 					shprintf("%s%s", p, sigtraps[i].name);
-			shprintf("\n");
+			shprintf(newline);
 		} else {
-			/* don't count signal 0, round up */
-			int nrows = (SIGNALS - 1 + 1) / 2;
-			int row, col;
+			int w, i;
+			int mess_width;
+			struct kill_info ki;
 
-			for (row = 0; row < nrows; row++)
-				for (col = 0; col < 2; col++) {
-					if (col)
-						shprintf("  ");
-					sig = row + col * nrows + 1;
-					if (sig >= SIGNALS)
-						break;
-					if (col == 0 && row != 0)
-						shprintf("\n");
-					t = &sigtraps[sig];
-					shprintf("%2d", t->signal);
-					if (t->name)
-						shprintf(" %8s", t->name);
-					else
-						shprintf(" %8d", t->signal);
-					shprintf(" %-26s", t->mess);
-				}
-			shprintf("\n");
+			for (i = SIGNALS, ki.num_width = 1; i >= 10; i /= 10)
+				ki.num_width++;
+			ki.name_width = mess_width = 0;
+			for (i = 0; i < SIGNALS; i++) {
+				w = sigtraps[i].name ? strlen(sigtraps[i].name)
+						     : ki.num_width;
+				if (w > ki.name_width)
+					ki.name_width = w;
+				w = strlen(sigtraps[i].mess);
+				if (w > mess_width)
+					mess_width = w;
+			}
+
+			print_columns(shl_stdout, SIGNALS - 1,
+				kill_fmt_entry, (void *) &ki,
+				ki.num_width + ki.name_width + mess_width + 3);
 		}
 		return 0;
 	}
@@ -842,13 +1044,16 @@ c_kill(wp)
 		if (*p == '%') {
 			if (j_kill(p, sig))
 				rv = 1;
+		} else if (!getn(p, &n)) {
+			bi_errorf("%s: arguments must be jobs or process ids",
+				p);
+			rv = 1;
 		} else {
-			n = getn(p);
 			/* use killpg if < -1 since -1 does special things for
 			 * some non-killpg-endowed kills
 			 */
 			if ((n < -1 ? killpg(-n, sig) : kill(n, sig)) < 0) {
-				shellf("kill: %s: %s\n", p, strerror(errno));
+				bi_errorf("%s: %s", p, strerror(errno));
 				rv = 1;
 			}
 		}
@@ -878,22 +1083,32 @@ c_getopts(wp)
 	char	*var;
 	int	optc;
 	char	buf[3];
+	struct tbl *vq;
 
-	ksh_getopt(wp, &builtin_opt, "");
+	if (ksh_getopt(wp, &builtin_opt, null) == '?')
+		return 1;
 	wp += builtin_opt.optind;
 
 	options = *wp++;
-	if (options == (char *) 0)
-		errorf("getopts: missing options argument\n");
+	if (options == (char *) 0) {
+		bi_errorf("missing options argument");
+		return 1;
+	}
 
 	var = *wp++;
-	if (var == (char *) 0)
-		errorf("getopts: missing name argument\n");
-	if (!*var || *skip_varname(var, TRUE))
-		errorf("getopts: %s: is not an identifier\n", var);
+	if (var == (char *) 0) {
+		bi_errorf("missing name argument");
+		return 1;
+	}
+	if (!*var || *skip_varname(var, TRUE)) {
+		bi_errorf("%s: is not an identifier", var);
+		return 1;
+	}
 
-	if (e->loc->next == (struct block *) 0)
-		errorf("getopts: internal error (no argv)\n");
+	if (e->loc->next == (struct block *) 0) {
+		internal_errorf(0, "c_getopts: no argv");
+		return 1;
+	}
 	/* Which arguments are we parsing... */
 	if (*wp == (char *) 0)
 		wp = e->loc->next->argv;
@@ -906,21 +1121,29 @@ c_getopts(wp)
 	if (user_opt.optind > argc
 	    || (user_opt.p != 0
 		&& user_opt.p > strlen(wp[user_opt.optind - 1])))
-	      errorf("getopts: arguments changed since last call\n");
+	{
+	      bi_errorf("arguments changed since last call");
+	      return 1;
+	}
 
 	user_opt.optarg = (char *) 0;
 	optc = ksh_getopt(wp, &user_opt, options);
 
-	if (optc >= 0 && (user_opt.info & GI_PLUS)) {
+	if (optc >= 0 && optc != '?' && (user_opt.info & GI_PLUS)) {
 		buf[0] = '+';
 		buf[1] = optc;
 		buf[2] = '\0';
 	} else {
+		/* POSIX says var is set to ? at end-of-options, at&t ksh
+		 * sets it to null - we go with POSIX...
+		 */
 		buf[0] = optc < 0 ? '?' : optc;
 		buf[1] = '\0';
 	}
-	/* todo: what if var is readonly */
-	setstr(global(var), buf);
+	vq = global(var);
+	if (vq->flag & RDONLY)
+		bi_errorf("%s is readonly", var);
+	setstr(vq, buf);
 
 	getopts_noset = 1;
 	setint(global("OPTIND"), (long) user_opt.optind);
@@ -942,29 +1165,35 @@ int
 c_bind(wp)
 	register char **wp;
 {
-	int macro = 0;
+	int rv = 0, macro = 0;
 	register char *cp;
+	int optc;
 
-	for (wp++; (cp = *wp) != NULL && *cp == '-'; wp++)
-		if (cp[1] == 'm')
+	while ((optc = ksh_getopt(wp, &builtin_opt, "m")) != EOF)
+		switch (optc) {
+		  case 'm':
 			macro = 1;
+			break;
+		  case '?':
+			return 1;
+		}
+	wp += builtin_opt.optind;
 
 	if (*wp == NULL)	/* list all */
-		x_bind((char*)NULL, (char*)NULL, 0);
+		rv = x_bind((char*)NULL, (char*)NULL, 0);
 
 	for (; *wp != NULL; wp++) {
 		cp = strchr(*wp, '=');
 		if (cp != NULL)
-			*cp++ = 0;
-		x_bind(*wp, cp, macro);
+			*cp++ = '\0';
+		if (x_bind(*wp, cp, macro))
+			rv = 1;
 	}
 
-	return 0;
+	return rv;
 }
 #endif
 
-extern	c_fc();
-
 /* A leading = means assignments before command are kept;
  * a leading * means a POSIX special builtin;
  * a leading + means a POSIX regular builtin
@@ -973,9 +1202,12 @@ extern	c_fc();
 const struct builtin kshbuiltins [] = {
 	{"+alias", c_alias},	/* no =: at&t manual wrong */
 	{"+cd", c_cd},
+	{"+command", c_command},
 	{"echo", c_print},
  	{"*=export", c_typeset},
+#ifdef HISTORY
 	{"+fc", c_fc},
+#endif /* HISTORY */
 	{"+getopts", c_getopts},
 	{"+jobs", c_jobs},
 	{"+kill", c_kill},
@@ -984,7 +1216,9 @@ const struct builtin kshbuiltins [] = {
  	{"*=readonly", c_typeset},
 	{"=typeset", c_typeset},
 	{"+unalias", c_unalias},
+#ifdef KSH
 	{"whence", c_whence},
+#endif /* KSH */
 #ifdef JOBS
 	{"+bg", c_fgbg},
 	{"+fg", c_fgbg},
